from setuptools import setup

setup(name='plantuml_adapter',
version='0.1',
description='MTC PLANTUML adapter',
url='',
author='',
author_email='',
license='',
packages=['plantuml_adapter'],
zip_safe=False)