from setuptools import setup


setup(name='datamodel',
      version='0.1',
      description='3SE datamodel',
      url='',
      author='',
      author_email='',
      license='',
      packages=['datamodel'],
      zip_safe=False)
